function [invR,V,Lambda,R]=CovarianceEstimatorSuite(X,method,param,verbose)

% Covariance Estimation Methods
% [V,Lambda]=CovarianceEstimatorSuite(X,method,param) returns a p-by-p
% precision matrix, inv(R), given jointly Gaussian iid samples.  Additional
% outputs include the estimated positive semi-definite covariance estimate,
% R, its eigenvectors, V, and p-by-1 list of non-negative eigenvalues,
% Lambda; thus, R=V*Lambda*V'.  Most techniques below adopt a model of
% R=M+v*I; the white noise variance, v, is alternatively referred to as
% "thermal noise power."  Techniques vary with respect to constraints
% assumed on M and on v.
%
% Inputs: 
%   X           p-by-n array of complex-valued observations, one sample per
%               column; assumed iid jointly Gaussian
%   method      string defining one of 15 methods below (default='scm')
%   param       parameter structure 'param' required in some methods      
%     .rank         rank
%     .var          variance (known thermal noise power)
%     .varLB        lower bound on thermal noise power
%     .varUB        upper bound on thermal noise power
%     .scale        p-by-p covariance (scale parameter) in InvWishart prior
%     .dof          degrees of freedom parameter in InvWishart prior
%   verbose     flag for showing messages to command window (default=0=off)
%       
% Outputs:
%   invR        inverse of the estimated covariance
%   V           p-by-p matrix of eigenvectors
%   Lambda      p-by-1 list of non-negative eigenvalues
%   R           estimated covariance matrix
%
% For a given application, computation may potentially be simplified to
% return only a desired subset of the outputs computed by this subroutine.
%
% Suite of Methods
% Maximum Likelihood Estimators & required parameters:
% 1 'scm'       Sample covariance matrix is the unconstrained MLE
% 2 'fml'       MLE for R=Rc+vI, v >= vmin (no rank assumed)
%               param.varLB; lower bound on thermal noise power.
% 3 'rank'      MLE for R=Rc+vI, rank(Rc)=param.rank.
%               param.rank, where param.rank < n
% 4 'rcml'      MLE for R=Rc+vI, rank(Rc)<=r, v known
%               param.rank, param.var
% 5 'rcml-lb'   MLE for R=Rc+vI, rank(Rc)<=rank<=n, v >= varLB
%               param.rank, param.varLB
% 6 'wax'       R=ABA'+vI, A p-by-r spans known subspace, v>0 unknown.
%               The estimator is not MLE.
%               param.A
% 7 'subspace'  MLE for R=ABA'+vI, A p-by-r (known range), v unknown >=0
%               param.A p-by-r spans a known subspace
% 8 'per'       MLE for R = Rc + vI, persymmetric, v >= varLB              
%               param.varLB
% 9 'rcper'     MLE for R = Rc +vI, persymmetric low-rank Rc
%               param.rank, param.varLB, param.varUB 
% 10 'cir'      MLE for R = Rc + vI, circulant
%               param.varLB
% 11 'rccir'    MLE for R = Rc +vI, circulant low-rank Rc
%               param.rank, param.varLB
%
% Maximum A Posteriori Probability Estimators & required parameters:
% 12 'map'      MAP estimator, inverse Wishart
%               param.dof, param.scale; dof>0 degrees of freedom, and scale
%               is a p-by-p covariance matrix for the prior
% 13 'rc-map'   MAP estimator for R=Rc+vI, low-rank Rc
%               param.rank,param.dof,param.scale,param.varLB,param.varUB
% 14 'per-map'	MAP for persymmetric R, inverse Wishart
%               param.dof, param.scale
% 15 'rcper-map' MAP for R=Rc+vI, low-rank persym., inverse Wishart
%               param.rank,param.dof,param.scale,param.varLB, param.varUB
%
% Example:
%     p=20;n=15;r=6;% 15 samples, 20 dimensions, rank 6
%     A = randn(p,r);%columns span rank r subspace
%     X = A*randn(r,n) + randn(p,n);%low rank plus thermal noise
%     param=struct('rank',r,'var',1);
%     [Rinv]=CovarianceEstimatorSuite(X,'rcml',param);
%
% References for methods are given in each case below.
%
% updated Mar 2019, LCP & AS
% v1.0
% Code should be referenced by citation to IEEE Signal Processing Letters,
% "Estimation of a Low-Rank Structured Covariance", by AS, LCP, and YC.
% 10.1109/LSP.2019.2906405
%
% Disclaimer: 
% Code provided as is with no warranty of fitness for any purpose.

% Notes: 
% (1) Matlab computations below are not necessarily optimized for speed.
% (2) while an MLE is best asymptotically normal, finite sample performance
% may be inferior to other estimators.
% (3) Results here assume iid AWGN per channel but may readily be extended.

% Read input arguments and set defaults
if(nargin<2); method='scm';end
if(nargin<4); verbose=0; end

switch lower(method)
    case {'scm','sample','sm'}
        %% 1. Sample Covariance Matrix
        % Sample covariance matrix is unconstrained MLE for iid zero-mean
        % jointly Gaussian observation vectors.
        [p,n]=size(X);%p dimensions, n samples (or "snapshots")
        if(verbose~=0)
            disp('Method: sample covariance matrix.')
            if(n<p)
                disp('    Warning: n less than p; return pinv(R)')
            end
        end
        % R=X*X'/n;%but, compute eigendecomposition to return variables
        [V,Lambda,~]=svd(X,'econ');% sorts singular values descending
        Lambda = diag((Lambda.^2)/n); %non-zero eigenvalues, sorted
        invR = V*diag((1./Lambda))*V';%note this is pseudo-inverse for n<p
        if(nargout == 4);R = V*diag(Lambda)*V';end
        
    case {'fml','steiner'}
        %% 2. FML 
        % M. Steiner and K. Gerlach, Fast converging adaptive processor 
        % for a structured covariance matrix, IEEE Trans. Aerosp. 
        % Electron. Syst., vol. 36, no. 4, pp. 11151126, 2000.
        % The mdoel is R = Rc + v*I, where Rc is clutter plus jammers, and 
        % v is iid thermal noise variance.  Assume only lower bound on v. 
        % No assumption regarding rank of Rc.  
        % Input parameter: param.varLB = noise floor
        [p,n]=size(X);%p dimensions, n samples (or "snapshots")
        if(exist('param','var')==0)
            error('FML method requires an input noise power, param.varLB.')
        elseif(isfield(param,'varLB')==0)
            error('FML method requires an input noise power, param.varLB.')
        elseif(param.varLB < 0)
            error('FML requires param.varLB to be non-negative.');
        end
        if(verbose~=0)
            fprintf('Method: FML, with thermal noise floor %g.\n',...
                param.varLB)
        end
        [V,Lambda,~]=svd(X); % singular value decomposition
        Lambda = diag((Lambda.^2)/n); %descending eigenvalues
        Lambda(Lambda<param.varLB) = param.varLB; %noise floor
        if(n < p); Lambda(n+1:p)=param.varLB;end
        invR = V*diag(1./Lambda)*V'; %inverse of covariance estimate
        if(nargout == 4);R = V*diag(Lambda)*V';end
 
    case {'rank', 'anderson'}
        %% 3. MLE for low rank plus unknown thermal noise
        % T. W. Anderson, Asymptotic theory for principal component
        % analysis, Ann. Math. Stat., vol. 34, no. 1, pp. 122148, 1963.
        % ML for low-rank clutter plus unknown thermal noise: R = v*I + Rc.
        % Given rank(Rc)=param.rank, this gives MLE for the two unknowns: 
        % noise floor, v, and positive semi-definite covariance, Rc.
        [p,n]=size(X);%p dimensions, n samples (or "snapshots")
        if(exist('param','var')==0)
            error('Method requires a clutter rank, param.rank.')
        elseif(isfield(param,'rank')==0)
            error('Method requires an input rank, param.rank.')
        elseif( round(param.rank)~=param.rank || param.rank < 1)
            error('Upper bound on rank must be integer greater than 0.')
        elseif( param.rank >= n )
            error('Method assumes rank < n.')
        end
        if(verbose~=0)
            disp('Method: rank-constrained MLE for R=Rc+v*I.')
            fprintf('    Rank of Rc bounded by %g.\n',param.rank)
        end
        [V,Lambda,~]=svd(X);
        Lambda = diag((Lambda.^2)/n); %descending abs values
        if(param.rank < p) %if constraint is present, else case is scm
            dbar = mean(Lambda(param.rank+1:end));%mean of noise space eigs
            Lambda = [Lambda(1:param.rank); dbar*ones(p-param.rank,1)];
        end
        invR = V*diag(1./Lambda)*V';
        if(nargout == 4);R = V*diag(Lambda)*V';end

    case {'rcml'}
        %% 4. Rank-constrained MLE, known thermal noise
        % B. Kang, V. Monga, and M. Rangaswamy, Rank-constrained maximum
        % likelihood estimation of structured covariance matrices, IEEE 
        % Trans. Aerosp. Electron. Syst., vol. 50, pp. 501515, 2014.
        % MLE given rank no greater than r, and known thermal noise power.
        [p,n]=size(X);%p dimensions, n samples (or "snapshots")
        if(exist('param','var')==0)
            error('Method requires parameter structure.')
        elseif(isfield(param,'rank')==0)
            error('Method requires an input rank, param.rank.')
        elseif(isfield(param,'var')==0)
            error('Method requires a known noise power, param.var.')
        elseif(param.rank < 1)
            error('Upper bound on rank must be 1 or greater.');
        end
        if(verbose~=0)
            disp('Method: rank-constrained ML with known noise power.')
            fprintf('    Clutter rank upper bounded by %g.\n',param.rank)
            fprintf('    Known noise power %g.\n',param.var)
        end
        [V,Lambda,~]=svd(X);
        Lambda = diag((Lambda.^2)/n); %descending abs value
        temp = min(n,param.rank);
        if(temp < p)
            Lambda(temp+1:p)=param.var;%known noise floor
        end
        Lambda(Lambda < param.var)=param.var;
        invR = V*diag(1./Lambda)*V';
        if(nargout == 4);R = V*diag(Lambda)*V';end
       
    case {'rcml-lb','rcml_lb'}       
        %% 5. Rank-constrained MLE with lower bound on thermal noise
        % B. Kang, V. Monga, and M. Rangaswamy, Rank-constrained maximum
        % likelihood estimation of structured covariance matrices, IEEE 
        % Trans. Aerosp. Electron. Syst., vol. 50, pp. 501515, 2014.
        % Given rank no greater than r for Rc, and lower bound, vmin, on v;
        % compute constrained minimization of likelihood for R=Rc+v*I.
        [p,n]=size(X);%p dimensions, n samples (or "snapshots")
        if(exist('param','var')==0)
            error('Method requires parameter structure.')
        elseif(isfield(param,'rank')==0)
            error('Method requires an input rank, param.rank.')
        elseif(isfield(param,'varLB')==0)
            error('Method requires a noise bound, param.varLB.')
        elseif(param.rank < 1)
            error('Upper bound on rank must be 1 or greater.');
        end
        if(verbose~=0)
            disp('Method: rank-constrained ML with lower bound on noise.')
            fprintf('    Clutter rank upper bounded by %g.\n',param.rank)
            fprintf('    Lower bound on noise power %g.\n',param.varLB)
        end        
        [V,Lambda,~]=svd(X);
        Lambda = diag((Lambda.^2)/n); %descending abs value
        %threshold mean of noise-space eigenvalues; keep e-vectors
        if(p > param.rank)
            if(n > param.rank)
                dbar=max(mean(Lambda(param.rank+1:min(p,n))),param.varLB);
                Lambda(param.rank+1:p)=dbar;
            elseif(n < p)% note: n <= rank
                Lambda(n+1:p) = param.varLB;dbar = param.varLB;
            end
        else
            dbar=param.varLB;
        end
        Lambda(Lambda < dbar)=dbar;
        invR = V*diag(1./Lambda)*V';
        if(nargout == 4);R = V*diag(Lambda)*V';end
         
    case {'wax'}       
        %% 6. Wax -- known signal subspace
        % Wax, Shan, and Kailath, Asilomar 1982 (as reported in Bresler 
        % 1988). Given A (and hence rank, r, also).  Gives MLE for the 
        % repeated noise eigenvalue, but ignores structure of known A other 
        % than rank.  Is, however, an unbiased, consistent, and 
        % positive semi-definite estimator of the covariance.
        [p,n]=size(X);%p dimensions, n samples (or "snapshots")
        if(exist('param','var')==0)
            error('Method requires parameter structure.')
        elseif(isfield(param,'A')==0)
            error('Method requires param.A spanning the signal space.')
        end
        A=param.A;r=size(A,2);pA=size(A,1);
        if( (r >= pA) || (pA ~= p) )
            error('Matrix size param.A is inappropriate.')
        elseif(n < p)
            error('Wax''s method assumes n >= p.')
        end
        if(verbose~=0)
            disp('Method: Wax''s subspace method, R=A*B*A''+v*I.')
            fprintf('    Columns in A: %g.\n',r)
        end
        [~,Lambda,~]=svd(X);
        Lambda = diag((Lambda.^2)/n); %descending abs value
        S = X*X'/n;%sample covariance matrix
        v_est = mean(Lambda(r+1:p)); %mean of noise eigenvalues
        P = pinv(A);
        R_wsk = v_est*eye(p) + A*P*(S - v_est*eye(p))*P'*A';
        %get eigen-decomposition
        [V,Lambda]=eig(R_wsk);
        Lambda=real(diag(Lambda));%symmetric, so eigs are real
        [Lambda,indx]=sort(Lambda,'descend');% sorted eigenvalues
        V=V(:,indx);%sort the eigenvectors
        invR = V*diag(1./Lambda)*V';
        if(nargout == 4);R = V*diag(Lambda)*V';end
        
    case {'subspace','yb','bresler'}       
    %% 7. Subspace MLE
        % Y. Bresler, "Maximum likelihood estimation of a linearly
        % structured covariance with application to antenna array
        % processing," Fourth Annual ASSP Workshop on Spectrum Estimation
        % and Modeling, Minneapolis, MN, USA, 1988, pp. 172-175. 
        % Known signal subspace, unknown noise floor.
        % Given A p-by-r and R=A*B*A'+v*I (and hence rank(A*B*A')<= r).
        % This is MLE for v and B, and is (almost) closed-form. 
        % Used in KB-GLRT by DeMaio, Farina, Wicks IEE-RSN 2005.
        [p,n]=size(X);%p dimensions, n samples (or "snapshots")
        if(exist('param','var')==0)
            error('Method requires parameter structure.')
        elseif(isfield(param,'A')==0)
            error('Method requires param.A spanning the signal space.')
        end
        A=param.A;r=size(A,2);pA=size(A,1);
        if( (r >= pA) || (pA ~= p) )
            error('Matrix size param.A is inappropriate.')
        end
        if(verbose~=0)
            disp('Method: Subspace MLE method, R=A*B*A''+v*I.')
            fprintf('    Columns in A: %g.\n',r)
        end
        S=X*X'/n;
        L=chol(A'*A,'lower');
        invL=inv(L);
        X=invL*A'*S*A*invL';
        [G,Phi]=eig(X); Phi=real(diag(Phi));%symmetric, so eigs are real
        [Phi,indx]=sort(Phi,'descend');G=G(:,indx);%sort the eigenvalues
        alpha=trace( (eye(p)-A*inv(A'*A)*A')*S );
        flag = 'true';
        r_est = r;
        Phi(r+1)=0; %just for first test in loop
        while(strcmp(flag,'true')) %search over rank
            if(r_est == r) 
                v_est=alpha/(p-r);
            else
                v_est = 1/(p-r_est)*(alpha + sum(Phi(r_est+1:end)));
            end
            if( (v_est >= Phi(r_est+1)) && (v_est <= Phi(r_est) ))
                flag='done';
            else
                r_est=r_est-1;
            end
        end
        Phi(r+1)=[];%remove dummy value used in first test in loop
        Phi = Phi-v_est; Phi(r_est+1:end)=0;
        R_yb = v_est*eye(p) + A*invL'*G*diag(Phi)*G'*invL*A';
        %get eigen-decomposition
        [V,Lambda]=eig(R_yb);
        Lambda=real(diag(Lambda));%Hermitian, so eigenvals are real
        [Lambda,indx]=sort(Lambda,'descend');% sorted eigenvalues
        V=V(:,indx);%sort the eigenvectors
        invR = V*diag(1./Lambda)*V';
        if(nargout == 4);R = V*diag(Lambda)*V';end
        
    case {'per'}        
        %% 8. Persymmetric ML with lower bound on thermal noise
        % Shikhaliev, Potter, Chi, "Estimation of a Low-Rank Structured 
        % Covariance", IEEE SP Letters, 2019.
        % Given lower bound on noise power, compute persymmetric MLE.
        % Recall persymmetric R is defined as R*J=J*R.', where
        % J=fliplr(eye(p)).
        [p,n]=size(X);%p dimensions, n samples (or "snapshots")
        if(exist('param','var')==0)
            error('Method requires parameter structure.')
        elseif(isfield(param,'varLB')==0)
            error('Method requires noise power lower bound, param.varLB.')
        end
        if(verbose~=0)
            disp('Method: persymmetric ML with noise power lower bound.')
            fprintf('    Lower bound on noise power %g.\n',param.varLB)
        end
        J = fliplr(eye(p));
        S = X*X'/n;%sample covariance matrix
        [V,Lambda,~] = svd(0.5*(S + J*conj(S)*J));%eig. for psd matrix
        Lambda=diag(Lambda);
        if(length(Lambda)<p) %n<p case
            Lambda(length(Lambda)+1:p)=param.varLB;
        end
        Lambda(Lambda < param.varLB) = param.varLB;%thermal noise floor
        invR = V*diag(1./Lambda)*V';
        if(nargout == 4);R = V*diag(Lambda)*V';end
        
    case {'rcper'}
        %% 9. Rank-constrained Persymmetric ML with noise bounds
        % Shikhaliev, Potter, Chi, "Estimation of a Low-Rank Structured
        % Covariance", IEEE SP Letters, 2019.
        % Given lower and upper bounds on noise power and upper bound on
        % rank, compute the constrained MLE.
        [p,n]=size(X);%p dimensions, n samples (or "snapshots")
        if(exist('param','var')==0)
            error('Method requires parameter structure.')
        elseif(isfield(param,'varLB')==0)
            error('Method requires noise power lower bound, param.varLB.')
        elseif(isfield(param,'varUB')==0)
            error('Method requires noise power upper bound, param.varUB.')
        elseif(isfield(param,'rank')==0)
            error('Method requires upper bound on rank, param.rank.')
        end
        if(verbose~=0)
            disp('Method: rank-constrained persymmetric ML')
            fprintf('    Lower bound on noise power %g.\n',param.varLB)
            fprintf('    Upper bound on noise power %g.\n',param.varUB)
            fprintf('    Clutter rank upper bounded by %g.\n',param.rank)
        end
        J = fliplr(eye(p));
        S = X*X'/n;%sample covariance matrix
        [V,Lambda,~] = svd(0.5*(S + J*conj(S)*J));%projection
        Lambda=diag(Lambda);
        if(param.rank < length(Lambda))%elseif, no noise space to average
            dbar =  mean(Lambda(param.rank+1:end));%noise space
            Lambda(param.rank+1:p) = min(param.varUB,dbar);
        elseif(length(Lambda) < p) %case of 2n <= min(rank,p)
            Lambda(length(Lambda)+1:p) = 0; %fill out set of eigs
        end
        Lambda(Lambda < param.varLB) = param.varLB;%thermal noise floor
        invR = V*diag(1./Lambda)*V';
        if(nargout == 4);R = V*diag(Lambda)*V';end
                  
    case {'cir'}
        %% 10. Circulant ML with lower bound on noise power
        % corrolary to Shikhaliev, Potter, Chi, "Estimation of a Low-Rank 
        % Structured Covariance", IEEE SP Letters, 2019.
        % Model is R=Rc+v*I, circulant Rc.
        % Given lower bound on noise power, compute MLE constrained to a
        % circulant covariance.
        [p,n]=size(X);%p dimensions, n samples (or "snapshots")
        if(exist('param','var')==0)
            error('Method requires parameter structure.')
        elseif(isfield(param,'varLB')==0)
            error('Method requires noise power lower bound, param.varLB.')
        end
        if(verbose~=0)
            disp('Method: circulant ML with noise bounds, R=Rc+v*I')
            fprintf('    Lower bound on noise power %g.\n',param.varLB)
        end
        F = 1/sqrt(p)*dftmtx(p);%unitary DFT matrix
        FX=F*X;
        Lambda = diag(FX*FX'/n);%projection of sample covariance
        [Lambda,indx]=sort(Lambda,'descend');%sort eigenvalues
        Lambda(Lambda < param.varLB) = param.varLB;%thermal noise floor
        V = (F(:,indx))';%sort eigenvectors
        invR = V*diag(1./Lambda)*V';
        if(nargout == 4);R = V*diag(Lambda)*V';end
        
    case {'rccir'}
        %% 11. Rank-constrained Circulant ML with noise bounds
        % corrolary to Shikhaliev, Potter, Chi, "Estimation of a Low-Rank 
        % Structured Covariance", IEEE SP Letters, 2019.  
        % Model is R=Rc+v*I, circulant Rc.
        % Given lower bound on noise power, v, compute MLE constrained to
        % a circulant covariance with low-rank Rc.
        [p,n]=size(X);%p dimensions, n samples (or "snapshots")
        if(exist('param','var')==0)
            error('Method requires parameter structure.')
        elseif(isfield(param,'varLB')==0)
            error('Method requires noise power lower bound, param.varLB.')
        elseif(isfield(param,'rank')==0)
            error('Must specify upper bound on clutter rank, param.rank.')
        elseif(param.rank >= p)
            error('Method expects param.rank < p.')
        end
        if(verbose~=0)
            disp('Method: circulant ML with noise bounds, R=Rc+v*I')
            fprintf('    Lower bound on noise power %g.\n',param.varLB)
            fprintf('    Clutter rank upper bounded by %g.\n',param.rank)
        end        
        F = 1/sqrt(p)*dftmtx(p);%unitary DFT
        FX=F*X;
        Lambda = diag(FX*FX'/n);%projection of sample covariance
        [Lambda,indx]=sort(Lambda,'descend');%sort eigenvalues
        Lambda(param.rank+1:end) = param.varLB;%noise subspace
        V = (F(:,indx))';%sort eigenvectors
        Lambda(Lambda < param.varLB) = param.varLB;%thermal noise floor
        invR = V*diag(1./Lambda)*V';
        if(nargout == 4);R = V*diag(Lambda)*V';end
        
    case {'map'}
        %% 12. MAP, InvWishart
        % Given InvWishart(L,R0) prior
        [p,n]=size(X);%p dimensions, n samples (or "snapshots")
        if(exist('param','var')==0)
            error('Method requires parameter structure.')
        elseif(isfield(param,'scale')==0)
            error('Method requires pd matrix, param.scale, for the prior.')
        elseif(isfield(param,'dof')==0)
            error('MAP requires InvWishart degrees of freedom, param.dof.')
        end
        if(verbose~=0)
            disp('Method: MAP with inverse Wishart prior')
        end
        R = (X*X' + param.scale);
        [V,Lambda]=svd(R);%eigenvectors
        Lambda = diag(Lambda)/(n+p+param.dof);%eigenvalues
        invR = V*diag(1./Lambda)*V';
        
    case {'rc-map','rcmap'}
        %% 13. Rank-constrained MAP, with noise power bounds
        % Shikhaliev, Potter, Chi, "Estimation of a Low-Rank Structured
        % Covariance", IEEE SP Letters, 2019. Given lower bound on noise
        % power, compute persymmetric MAP with constrained InvWishart
        % prior. Recall persymmetric R is defined as R*J=J*R.', where
        % J=fliplr(eye(p)).
        [p,n]=size(X);%p dimensions, n samples (or "snapshots")
        if(exist('param','var')==0)
            error('Method requires parameter structure.')
        elseif(isfield(param,'scale')==0)
        elseif(isfield(param,'scale')==0)
            error('Method requires pd matrix, param.scale, for the prior.')
        elseif(isfield(param,'dof')==0)
            error('MAP requires InvWishart degrees of freedom, param.dof.')
        elseif(isfield(param,'varLB')==0)
            error('Method requires noise power lower bound, param.varLB.')
        elseif(isfield(param,'varUB')==0)
            error('Method requires noise power upper bound, param.varUB.')
        elseif(isfield(param,'rank')==0)
            error('Method requires upper bound on rank, param.rank.')
        end
        if(verbose~=0)
            disp('Method: low-rank MAP with inverse Wishart prior')
            fprintf('    Lower bound on noise power %g.\n',param.varLB)
            fprintf('    Upper bound on noise power %g.\n',param.varUB)
            fprintf('    Clutter rank upper bounded by %g.\n',param.rank)
        end
        [V,Lambda]=svd( X*X' + param.scale );
        Lambda = diag(Lambda)/(n+p+param.dof);    
        if(param.rank < p)%if rank>=p, inactive
            dbar =  mean(Lambda(param.rank+1:end));%noise space
            Lambda(param.rank+1:p) = min(param.varUB,dbar);
        end
        Lambda(Lambda < param.varLB) = param.varLB;%thermal noise floor
        invR = V*diag(1./Lambda)*V';
        if(nargout == 4);R = V*diag(Lambda)*V';end

    case {'per-map','permap'}
        %% 14. Persymmetric MAP with noise floor
        % Shikhaliev, Potter, Chi, "Estimation of a Low-Rank Structured
        % Covariance", IEEE SP Letters, 2019. Given lower bound on noise
        % power, compute persymmetric MAP with constrained 
        % InvWishart prior. Recall persymmetric R is defined as R*J=J*R.', 
        % where J=fliplr(eye(p)).
        [p,n]=size(X);%p dimensions, n samples (or "snapshots")
        if(exist('param','var')==0)
            error('Method requires parameter structure.')
        elseif(isfield(param,'scale')==0)
            error('Method requires a matrix, param.scale for the prior.')
        elseif(isfield(param,'dof')==0)
            error('Method requires Wishart degrees of freedom, param.dof.')
        elseif(isfield(param,'varLB')==0)
            error('Method requires noise power lower bound, param.varLB.')
        end
        if(verbose~=0)
            disp('Method: persymmetric MAP with InvWishart prior')
            fprintf('    Lower bound on noise power %g.\n',param.varLB)
        end
        Rmap=(X*X' + param.scale);%MAP
        J=fliplr(eye(p));
        [V,Lambda] = svd(0.5*(Rmap + J*conj(Rmap)*J));%translate & project
        Lambda = diag(Lambda)/(n+p+param.dof);  %eigenvalues  
        Lambda(Lambda < param.varLB) = param.varLB;%thermal noise floor
        invR = V*diag(1./Lambda)*V';
        if(nargout == 4);R = V*diag(Lambda)*V';end
                  
    case {'rcper-map','rc-per-map'}
        %% 15. Rank-constrained persymmetric MAP with noise bounds
        % Shikhaliev, Potter, Chi, "Estimation of a Low-Rank Structured
        % Covariance", IEEE SP Letters, 2019. Given lower bound on noise
        % power, compute persymmetric MAP of low-rank with constrained 
        % InvWishart prior. Recall persymmetric R is defined as R*J=J*R.', 
        % where J=fliplr(eye(p)).
        [p,n]=size(X);%p dimensions, n samples (or "snapshots")
        if(exist('param','var')==0)
            error('Method requires parameter structure.')
        elseif(isfield(param,'scale')==0)
            error('Method requires a matrix, param.scale for the prior.')
        elseif(isfield(param,'dof')==0)
            error('Method requires Wishart degrees of freedom, param.dof.')
        elseif(isfield(param,'varLB')==0)
            error('Method requires noise power lower bound, param.varLB.')
        elseif(isfield(param,'varUB')==0)
            error('Method requires noise power upper bound, param.varUB.')
        elseif(isfield(param,'rank')==0)
            error('Method requires upper bound on rank, param.rank.')
        end
        if(verbose~=0)
            disp('Method: low-rank persymmetric MAP with InvWishart prior')
            fprintf('    Lower bound on noise power %g.\n',param.varLB)
            fprintf('    Upper bound on noise power %g.\n',param.varUB)
            fprintf('    Clutter rank upper bounded by %g.\n',param.rank)
        end
        Rmap=(X*X' + param.scale);%MAP
        J=fliplr(eye(p));
        [V,Lambda] = svd(0.5*(Rmap + J*conj(Rmap)*J));%project to persym. 
        Lambda = diag(Lambda)/(n+p+param.dof);%eigenvalues
        if(param.rank < p)%if rank>=p, inactive
            dbar =  mean(Lambda(param.rank+1:end));%noise space
            Lambda(param.rank+1:p) = min(param.varUB,dbar);
        end
        Lambda(Lambda < param.varLB) = param.varLB;%thermal noise floor
        invR = V*diag(1./Lambda)*V';
        if(nargout == 4);R = V*diag(Lambda)*V';end
               
    otherwise
        %% none of the above
        disp('Unknown method.')
end
