% Script to create figure 1(a) in Shikhaliev, Potter, Chi, "Estimation of a
% Low-Rank Structured Covariance", IEEE SP Letters, 2019.
% doi 10.1109/LSP.2019.2906405
%
% Last revised, Mar 2019, AS
% Disclaimer: 
% Code provided as is with no warranty of fitness for any purpose.

close all;clear variables;
tic; 
%% Create a structured low-rank + scaled identity matrix
mat_size = 16; 
F = 1/sqrt(mat_size)*dftmtx(mat_size);
clut_pow = 10;
therm_noise_pow = 1;
low_rank = 5;
MCsim = 10000;%number of MC trials

rho = 0.9; %Used in creating the true covariance matrix in an exponential 
           %tapering form.     
Rrho = toeplitz(rho.^(0:mat_size-1).');       

[U,S] = svd(Rrho);
R0 = U(:,1:low_rank)*S(1:low_rank,1:low_rank)*U(:,1:low_rank)' + ...
   therm_noise_pow*eye(mat_size);
R0 = R0 - diag(diag(R0)) + real(diag(diag(R0)));
P0 = inv(R0); 

[U,S] = svd(R0);
R0_sqrt = U*diag(sqrt(diag(S)));                                       
  
svec = 1/sqrt(mat_size)*ones(mat_size,1); %Steering vector

%% Simulate at various numbers of samples ("snapshots")
snapLevels = 8:4:44;
num_snapLevels = length(snapLevels);
SINR_at_each_snapLevel_Pmap = zeros(1,num_snapLevels);
SINR_at_each_snapLevel_Prcmap = zeros(1,num_snapLevels);
SINR_at_each_snapLevel_Ppsm_map = zeros(1,num_snapLevels);
SINR_at_each_snapLevel_Prcpsm_map = zeros(1,num_snapLevels);

J = fliplr(eye(mat_size));
for k1 = 1:num_snapLevels
    Nsnaps = snapLevels(k1);
    for k2 = 1:MCsim
        str1 = ['Monte-Carlo Simulation Number: ',num2str(k2), ' out of ', ...
            num2str(MCsim), ' for range bin ',num2str(k1), ' out of ',...
            num2str(num_snapLevels)];
        disp(str1);    

        %"R0" is the prior and mean matrix parameterizing the Wishart
        %distribution from which the true covariance matrix "Rtrue" is sampled
        %from.
        DOF = mat_size+1; % Degrees of freedom for the Wishart distribution
        z_snaps = 1/sqrt(2)*R0_sqrt*(randn(mat_size,DOF) + 1i*randn(mat_size,DOF));
        Rtrue = 1/DOF*z_snaps*(z_snaps');

        %Enforcing rank and structural constraints on "Rtrue":
        Rtrue = 0.5*(Rtrue + J*conj(Rtrue)*J); %make persymmetric
        Rtrue = A_func_noise_floor(Rtrue,low_rank,therm_noise_pow);
        Rtrue = Rtrue - diag(diag(Rtrue)) + real(diag(diag(Rtrue)));

        Ptrue = inv(Rtrue);
        [U,S] = svd(Rtrue);
        Rtrue_sqrt = U*diag(sqrt(diag(S))); %Gave more numerically accurate results 
                                    %than the built-in "sqrt()" and "chol()"
        
        %Generate secondary data 
        snaps = Rtrue_sqrt*1/sqrt(2)*(randn(mat_size,Nsnaps) + ...
            1i*randn(mat_size,Nsnaps));                
        
        Rscm = 1/size(snaps,2)*(snaps)*snaps';
        Rmap = 1/(Nsnaps + mat_size + DOF)*(Nsnaps*Rscm + DOF*R0);
        Pmap = inv(Rmap);
        
        Rrcmap = A_func_noise_floor(Rmap,low_rank,therm_noise_pow);
        Prcmap = inv(Rrcmap);

        Rpsm_map = 0.5*(Rmap + J*conj(Rmap)*J);
        Ppsm_map = inv(Rpsm_map);

        Rrcpsm_map = A_func_noise_floor(Rpsm_map,low_rank,therm_noise_pow);
        Prcpsm_map = inv(Rrcpsm_map);     

        %Finding the SINRs
        SINR_at_each_snapLevel_Pmap(k1) = SINR_at_each_snapLevel_Pmap(k1) +...
            abs(svec'*Pmap*svec)^2/(abs(svec'*Pmap*Rtrue*Pmap*svec)*...
            abs(svec'*Ptrue*svec));
        
        SINR_at_each_snapLevel_Ppsm_map(k1) = SINR_at_each_snapLevel_Ppsm_map(k1) +...
            abs(svec'*Ppsm_map*svec)^2/(abs(svec'*Ppsm_map*Rtrue*Ppsm_map*svec)*...
            abs(svec'*Ptrue*svec));        
        
        SINR_at_each_snapLevel_Prcpsm_map(k1) = SINR_at_each_snapLevel_Prcpsm_map(k1) +...
            abs(svec'*Prcpsm_map*svec)^2/(abs(svec'*Prcpsm_map*Rtrue*Prcpsm_map*svec)*...
            abs(svec'*Ptrue*svec));          
        
        SINR_at_each_snapLevel_Prcmap(k1) = SINR_at_each_snapLevel_Prcmap(k1) +... 
            abs(svec'*Prcmap*svec)^2/(abs(svec'*Prcmap*Rtrue*Prcmap*svec)*...
            abs(svec'*Ptrue*svec));              
    end
end

%Normalizing the Pd at each SINR
SINR_at_each_snapLevel_Pmap = SINR_at_each_snapLevel_Pmap/MCsim;
SINR_at_each_snapLevel_Prcmap = SINR_at_each_snapLevel_Prcmap/MCsim;
SINR_at_each_snapLevel_Ppsm_map = SINR_at_each_snapLevel_Ppsm_map/MCsim;
SINR_at_each_snapLevel_Prcpsm_map = SINR_at_each_snapLevel_Prcpsm_map/MCsim;

%% Figure
figure;
plot(snapLevels,10*log10(SINR_at_each_snapLevel_Pmap),'-p','MarkerSize',5); 
hold on;
plot(snapLevels,10*log10(SINR_at_each_snapLevel_Prcmap),'->','MarkerSize',5);  
plot(snapLevels,10*log10(SINR_at_each_snapLevel_Ppsm_map),'-s','MarkerSize',5); 
plot(snapLevels,10*log10(SINR_at_each_snapLevel_Prcpsm_map),'-^','MarkerSize',5); 

xlabel('Number of Snapshots');
ylabel('Normalized SINR (dB)');
legend('MAP','RC-MAP','PSM-MAP','RC-PSM-MAP','Location','southeast');
title('SINR Across Snapshots');
grid
toc
%end of routine