% Script to create figure 1(b) in Shikhaliev, Potter, Chi, "Estimation of a
% Low-Rank Structured Covariance", IEEE SP Letters, 2019.
% doi 10.1109/LSP.2019.2906405
%
% Last revised, Mar 2019, AS
% Disclaimer: 
% Code provided as is with no warranty of fitness for any purpose.

close all;clear variables;
tic;
%% Create a structured low-rank + scaled identity matrix
mat_size = 16; 
F = 1/sqrt(mat_size)*dftmtx(mat_size);
clut_pow = 10;
therm_noise_pow = 1;
low_rank = 5;
MCsim = 10000;%number of MC trials 

rho = 0.9; %Used in creating the true covariance matrix in an exponential 
           %tapering form         
Rrho = toeplitz(rho.^(0:mat_size-1).');         
[U,S] = svd(Rrho);
Rtrue = U(:,1:low_rank)*S(1:low_rank,1:low_rank)*U(:,1:low_rank)' + ...
   therm_noise_pow*eye(mat_size);
Rtrue = Rtrue - diag(diag(Rtrue)) + real(diag(diag(Rtrue)));
Ptrue = inv(Rtrue); 
[U,S] = svd(Rtrue);
Rtrue_sqrt = U*diag(sqrt(diag(S))); %Gave more numerically accurate results 
                                    %than the built-in "sqrt()" and "chol()"
                                        
svec = 1/sqrt(mat_size)*ones(mat_size,1); %Steering vector

%% Simulate at various numbers of samples ("snapshots")
snapLevels = 8:4:44;  
num_snapLevels = length(snapLevels);

SINR_at_each_snapLevel_Pscm = zeros(1,num_snapLevels);
SINR_at_each_snapLevel_Prcml = zeros(1,num_snapLevels);
SINR_at_each_snapLevel_Ppsm = zeros(1,num_snapLevels);
SINR_at_each_snapLevel_Prcpsm = zeros(1,num_snapLevels);

J = fliplr(eye(mat_size));
for k1 = 1:num_snapLevels
    Nsnaps = snapLevels(k1);
    for k2 = 1:MCsim
        str1 = ['Monte-Carlo Simulation Number: ',num2str(k2), ' out of ', ...
            num2str(MCsim), ' for range bin ',num2str(k1), ' out of ',...
            num2str(num_snapLevels)];
        disp(str1);    

        %Generate secondary data (target-free)
        snaps = Rtrue_sqrt*1/sqrt(2)*(randn(mat_size,Nsnaps) + ...
            1i*randn(mat_size,Nsnaps));
        
        Rscm = 1/size(snaps,2)*(snaps)*snaps';
        Pscm = pinv(Rscm);
        
        Rrcml = A_func_noise_floor(Rscm,low_rank,therm_noise_pow);
        Prcml = inv(Rrcml);

        Rpsm = 0.5*(Rscm + J*conj(Rscm)*J);
        Ppsm = pinv(Rpsm);

        Rrcpsm = A_func_noise_floor(Rpsm,low_rank,therm_noise_pow);
        Prcpsm = inv(Rrcpsm);        

        %Finding the SINRs      
        if (Nsnaps>=size(Rscm,1))
        SINR_at_each_snapLevel_Pscm(k1) = SINR_at_each_snapLevel_Pscm(k1) +...
            abs(svec'*Pscm*svec)^2/(abs(svec'*Pscm*Rtrue*Pscm*svec)*...
            abs(svec'*Ptrue*svec));
        else
            SINR_at_each_snapLevel_Pscm(k1) = 0;
        end
        
        SINR_at_each_snapLevel_Ppsm(k1) = SINR_at_each_snapLevel_Ppsm(k1) +...
            abs(svec'*Ppsm*svec)^2/(abs(svec'*Ppsm*Rtrue*Ppsm*svec)*...
            abs(svec'*Ptrue*svec));    
        
        SINR_at_each_snapLevel_Prcpsm(k1) = SINR_at_each_snapLevel_Prcpsm(k1) +...
            abs(svec'*Prcpsm*svec)^2/(abs(svec'*Prcpsm*Rtrue*Prcpsm*svec)*...
            abs(svec'*Ptrue*svec));          
        
        SINR_at_each_snapLevel_Prcml(k1) = SINR_at_each_snapLevel_Prcml(k1) +... 
            abs(svec'*Prcml*svec)^2/(abs(svec'*Prcml*Rtrue*Prcml*svec)*...
            abs(svec'*Ptrue*svec));        
        
    end
end

%Normalizing the Pd at each SINR
SINR_at_each_snapLevel_Pscm = SINR_at_each_snapLevel_Pscm/MCsim;
SINR_at_each_snapLevel_Prcml = SINR_at_each_snapLevel_Prcml/MCsim;
SINR_at_each_snapLevel_Ppsm = SINR_at_each_snapLevel_Ppsm/MCsim;
SINR_at_each_snapLevel_Prcpsm = SINR_at_each_snapLevel_Prcpsm/MCsim;

%% Figure
figure;
plot(snapLevels(3:end),10*log10(SINR_at_each_snapLevel_Pscm(3:end)), ...
    '-p','MarkerSize',5); 
hold on;
plot(snapLevels,10*log10(SINR_at_each_snapLevel_Prcml),'->','MarkerSize',5);  
plot(snapLevels,10*log10(SINR_at_each_snapLevel_Ppsm),'-s','MarkerSize',5); 
plot(snapLevels,10*log10(SINR_at_each_snapLevel_Prcpsm),'-^','MarkerSize',5); 

xlabel('Number of Snapshots');
ylabel('Normalized SINR (dB)');
legend('ML','RC-ML','PSM-ML','RC-PSM-ML','Location','southeast');
title('SINR Across Snapshots');
grid
toc
%end of routine

